# xDevs.com Calibration app for Keithley 2000, v0.2
# https://xdevs.com/fix/kei2400
# https://xdevs.com/fix/hp3458a
import sys
import time
import Gpib
import signal
import numbers
from math import log10

dmm_val = 1.0
dmm_temp = 36.6

class Timeout():
  """Timeout class using ALARM signal"""
  class Timeout(Exception): pass

  def __init__(self, sec):
    self.sec = sec

  def __enter__(self):
    signal.signal(signal.SIGALRM, self.raise_timeout)
    signal.alarm(self.sec)

  def __exit__(self, *args):
    signal.alarm(0) # disable alarm

  def raise_timeout(self, *args):
    raise Timeout.Timeout()

class dmm():
    global dmm_val 
    data = ""
    temp = 36.6
    cnti = 12
    
    def __init__(self,gpib,refhp,name):
        self.gpib = gpib
        self.inst = Gpib.Gpib(0,self.gpib, timeout=60) # 3458A GPIB Address = self.gpib
        self.refhp = refhp
        self.name = name
        self.init_inst()

    def init_inst(self):
        # Setup HP 3458A
        self.inst.write ("PRESET NORM")
        self.inst.write ("OFORMAT ASCII")
        self.inst.write ("FUNC DCV,AUTO")
        self.inst.write ("TARM HOLD")
        self.inst.write ("TRIG LINE")
        self.inst.write ("NPLC 50")
        self.inst.write ("AZERO ON")
        self.inst.write ("NRDGS 1,AUTO")
        self.inst.write ("MEM OFF")
        self.inst.write ("END ALWAYS")
        self.inst.write ("NDIG 9")

    def switch_dci(self):
        self.inst.write ("PRESET NORM")
        self.inst.write ("NPLC 50")
        self.inst.write ("AZERO ON")
        self.inst.write ("NRDGS 1,AUTO")
        self.inst.write ("OFORMAT ASCII")
	self.inst.write ("FUNC DCI,AUTO")
        self.inst.write ("NDIG 9")

    def switch_ohmf(self):
        self.inst.write ("PRESET NORM")
        self.inst.write ("NPLC 50")
        self.inst.write ("AZERO ON")
        self.inst.write ("NRDGS 1,AUTO")
        self.inst.write ("OFORMAT ASCII")
	self.inst.write ("FUNC OHMF,AUTO")
        self.inst.write ("NDIG 9")
        
    def read_data(self,cmd):
        data_float = 0.0
        data_str = ""
        self.inst.write(cmd)
        
        try:
            with Timeout(20):
                data_str = self.inst.read()
        except Timeout.Timeout:
            print ("Timeout exception from dmm %s on read_data() inst.read()\n" % self.name)
            return (0,float(0))
        #print ("Reading from dmm %s = %s" % (self.name,data_str))
        try:
            data_float = float(data_str)
        except ValueError:
            print("Exception thrown by dmm %s on read_data() - ValueError = %s\n" % (self.name,data_str))
            return (0,float(0)) # Exception on float conversion, 0 = error
        return (1,data_float) # Good read, 1 = converted to float w/o exception
        
    def get_temp(self):
        global dmm_temp
        print("Reading DMM temp")
        self.inst.write("TARM SGL,1")
        self.temp_status_flag,temp = self.read_data("TEMP?")
        if (self.temp_status_flag):
            self.temp = temp
	    dmm_temp = temp
        return self.temp
        
    def get_temp_status(self):
        return self.temp_status_flag

    def get_data(self):
        global dmm_val
        self.status_flag,data = self.read_data("TARM SGL,1")
        if (self.status_flag):
            self.data = data
            dmm_val = float(data)
            self.ppm = ((float(self.data) / self.refhp)-1)*1E6
        return self.data
        
    def exec_acal(self):
        sys.stdout.write ("\033[1;35mACAL ALL procedure start, please wait 12 minutes ")
        self.inst.write("ACAL ALL")
        for cnt in range(0 ,48):        # wait 720 seconds, print dot every 15s
            sys.stdout.write ("\033[0;43m*")
            time.sleep(15)
        print("\r\nACAL procedure done...\033[1;39m")
    
    def exec_idn(self):
        self.inst.write ("END ALWAYS")
        self.inst.write ("ID?")
	dat = self.inst.read()
	tstr = dat.split()
        if (tstr[0] == "HP3458A"):
            sys.stdout.write ("\r\n\033[1;32m%s detected...\033[1;39m" % tstr[0])
        else:
            sys.stdout.write ("\r\n\033[1;31mNo DMM present, exiting!\033[1;39m")
            quit()

    def get_data_status(self):
        return self.status_flag

class mfc():
    global dmm_val 
    data = ""
    temp = 36.6
    cnti = 12
    
    def __init__(self,gpib,refhp,name):
        self.gpib = gpib
        self.inst = Gpib.Gpib(0,self.gpib, timeout=60) # 3458A GPIB Address = self.gpib
        self.refhp = refhp
        self.name = name
        self.init_inst()

    def init_inst(self):
        # Setup MFC
	print ("No init done here\r\n")
        #self.inst.write ("E1\n")

    def dcv(self,voltage):
	print ("Setting MFC to %f\r\n" % float(voltage))
        self.inst.write ("%f\n" % voltage)

    def ohmf(self,res):
	res_mfc_id = [1, 2, 3, 4, 5, 6, 7]
	res_mfc = ["10 Ohm","100 Ohm","1 KOhm", "10 KOhm", "100 KOhm", "1 MOhm", "10 MOhm"]
	
	if ((res > 0) and (res < 8)):
	    print ("Setting MFC to %d R\r\n" % res_mfc[res] )
    	    self.inst.write ("O%c\n" % res_mfc_id[res])
	else:
	    print ("Incorrect MFC ohm setting, valid from 1 to 7\r\n")
        
    def read_data(self,cmd):
        data_float = 0.0
        data_str = ""
        self.inst.write(cmd)
        
        try:
            with Timeout(20):
                data_str = self.inst.read()
        except Timeout.Timeout:
            print ("Timeout exception from dmm %s on read_data() inst.read()\n" % self.name)
            return (0,float(0))
        #print ("Reading from dmm %s = %s" % (self.name,data_str))
        try:
            data_float = float(data_str)
        except ValueError:
            print("Exception thrown by dmm %s on read_data() - ValueError = %s\n" % (self.name,data_str))
            return (0,float(0)) # Exception on float conversion, 0 = error
        return (1,data_float) # Good read, 1 = converted to float w/o exception
        
    def get_temp_status(self):
        return self.temp_status_flag

    def get_data(self):
        global dmm_val
        self.status_flag,data = self.read_data("TARM SGL,1")
        if (self.status_flag):
            self.data = data
            dmm_val = float(data)
            self.ppm = ((float(self.data) / self.refhp)-1)*1E6
        return self.data
        
    def exec_idn(self):
        self.inst.write ("END ALWAYS")
        self.inst.write ("ID?")
	dat = self.inst.read()
	tstr = dat.split()
        if (tstr[0] == "HP3458A"):
            sys.stdout.write ("\r\n\033[1;32m%s detected...\033[1;39m" % tstr[0])
        else:
            sys.stdout.write ("\r\n\033[1;31mNo DMM present, exiting!\033[1;39m")
            quit()

    def get_data_status(self):
        return self.status_flag


with open('callog_ver.txt','wb') as b:
    kei2000 = Gpib.Gpib(0,20)
    kei2000.clear()

    ref_dci_set = [100E-6, 1E-3, 1E-2, 1E-1, 1.0]
    ref_dci_meas = [100E-6, 1E-3, 1E-2, 1E-1, 1.0]
    cal_dcv_set = [-0.2, 0, 0.2, 0, -2, 0, 2, 0, -20, 0, 20, 0, -200, 0, 200, 0]
    cal_dcv_rng = [0.2, 0.2, 0.2, 0.2, 2, 2, 2, 2, 20, 20, 20, 20, 200, 200, 200, 200]
    cal_dcv_str = ["-0.2 m"," -Zero"," 0.2 m"," +Zero","   -2 "," -Zero","    2 "," -Zero","  -20 "," +Zero","   20 "," -Zero","\033[0;41m -200 \033[0;39m"," +Zero","\033[0;41m  200 \033[0;39m"," +Zero"]
    cal_dci_set = [-1E-6, 0, 1E-6, 0, -10E-6, 0, 10E-6, 0, -100E-6, 0, 100E-6, 0, -1E-3, 0, 1E-3, 0, -10E-3, 0, 10E-3, 0, -100E-3, 0, 100E-3, 0, -1, 0, 1, 0]
    cal_dci_rng = [1E-6, 1E-6, 1E-6, 1E-6, 10E-6, 10E-6, 10E-6, 10E-6, 100E-6, 100E-6, 100E-6, 100E-6, 1E-3, 1E-3, 1E-3, 1E-3, 10E-3, 10E-3, 10E-3, 10E-3, 100E-3, 100E-3, 100E-3, 100E-3, 1, 1, 1, 1]
    cal_dci_str = ["-1 u","-Zero", " 1 u", "+Zero", " -10 u", "-Zero", " 10 u", "+Zero", "-100 u", "-Zero", " 100 u", "+Zero", "-1 m", "-Zero", " 1 m", "+Zero", " -10 m", "-Zero", " 10 m", "+Zero", "-100 m", "-Zero", " 100 m", "+Zero", " -1 ", "-Zero", " 1 ", "+Zero"]
    test_dcv_set = [0, -0.002, 0.002, -0.02, 0.02, -0.2, 0.2, -1, 1, -2, 2, -10, 10, -15, 15, -20, 20, -100, 100, -200, 200, 0]
    test_dcv_rng = [0.2,  0.2,   0.2,   0.2,  0.2,  0.2, 0.2,  2, 2, 20,20,  20, 20,  20, 20,  20, 20, 200, 200, 200, 200, 20]
    test_dci_set = [0, -1E-9, 1E-9, -10E-9, 10E-9, -100E-9, 100E-9, -1E-6, 1E-6, -10E-6, 10E-6, -100E-6, 100E-6, -1E-3, 1E-3, -4E-3, 4E-3, -10E-3, 10E-3, -20E-3, 20E-3, -100E-3, 100E-3, -500E-3, 500E-3, 1, -1, 0]
    test_dci_rng = [1E-6, 1E-6, 1E-6, 1E-6, 1E-6, 1E-6,  1E-6, 1E-6, 1E-6, 10E-6, 10E-6, 100E-6, 100E-6, 1E-3, 1E-3, 10E-3, 10E-3, 10E-3, 10E-3, 100E-3, 100E-3, 100E-3, 100E-3, 1, 1, 1, 1, 1E-3]
    test_dci_str = ["Zero  ","  -1 n", "   1 n", " -10 n", "  10 n", "-100 n", " 100 n", "  -1 u", "   1 u", " -10 u", "  10 u", "-100 u", " 100 u", "  -1 m", "   1 m", "  -4 m", "   4 m", " -10 m", "  10 m", " -20 m", "  20 m", "-100 m", " 100 m", "-500 m", " 500 m", "  -1  ", "   1  ", "Zero  "]
    kei2000_idn = "KEITHLEY INSTRUMENTS INC."
    kei2000_mdl = "MODEL 2000"
    orig_val = []
    kei_error = 0
    plc = 60 #60 Hz
    cnt = 0
    samples = 0
    nplc = 0
    val = 3.423
    meas = 0.0
    temp = 36.6
    
    for cnt in range(30):
        orig_val.append(0)
        
    sys.stdout.write("\033[0;36m Keithley 2000 calibration tool \r\n  Using NI GPIB adapter with next instruments config: \r\n * GPIB 20 : Keithley 2000 DMM \r\n * GPIB 3 : HP 3458A 8.5-digit DMM \r\n * GPIB 21 : Time 9823 MFC \r\n ! Do not swap terminals during calibration, high voltage may be present !\033[0;39m\r\n")
    
    kei2000.write ("*IDN?")
    idn_str = kei2000.read()
    print idn_str
    idmfg = idn_str.split(",")
    if (kei2000_idn == idmfg[0]):
        if (kei2000_mdl == idmfg[1]):
            sys.stdout.write("\033[0;32mKeithley Model 2000 - detected, S/N %s, Version: %s\033[0;39m\r\n" % (idmfg[2], idmfg[3]))
	    b.write("\r\nKeithley Model 2000 - detected, S/N %s, Version: %s\r\n" % (idmfg[2], idmfg[3]))
        else:
            print ("\033[0;31mIncorrect Model! Check GPIB address. Calibration cancelled.\033[0;39m")    
            quit()
    else:
        sys.stdout.write ("\033[0;31mNo Keithley instrument detected. Check GPIB address. Calibration cancelled.\033[0;39m")
        quit()
    
    kei2000.write (":SYST:PRES;*CLS")
    kei2000.write ("*ESE 1;*SRE 32")
    
    #kei2000.write (":CAL:PROT:LOCK")
    #kei2000.write (":CAL:PROT:CODE 'KI002000'")
    
    dmm = dmm(3,dmm_val,"3458A") # GPIB 3
    dmm.exec_idn()
    dmm.get_temp()
    
    print ("\033[0;33mHP3458A TEMP = %2.1f C\033[0;33m" % dmm_temp)
    b.write ("HP 3458A detected\r\n TEMP? = %2.1f C\r\n" % dmm_temp)

    mfc = mfc(21,dmm_val,"3458A") # GPIB 3
    mfc.inst.write("R4\n") # set range 10.0
    mfc.dcv(0.000000) # output zero DCV

    print ("\033[1;31mConnect 3458A DMM volts input to Model 2000 INPUT/OUTPUT jacks and to MFC.\033[1;39m")
    raw_input("\033[1;33mPress Enter to continue with DCV test...\033[1;39m")
    
    #dmm.exec_acal()
    
    kei2000.write (":SENS:FUNC 'VOLT:DC'")
    kei2000.write (":SENS:VOLT:DC:DIG 7;NPLC 10;AVER:COUN 5; TCON REP")
    kei2000.write (":SENS:VOLT:DC:AVER:STAT ON")
    kei2000.write (":SOUR:VOLT:DC:RANG 10")
    kei2000.write (":SYST:BEEP:STAT OFF")
    kei2000.write (":TRIG:COUN 1")
    kei2000.write (":TRIG:SOUR IMM")
    kei2000.write (":FORM:ELEM READ")

    print ("\033[1;31mConnect output current terminals at MFC to 3458A Reference DMM to record currents \033[1;39m")
    raw_input("\033[1;33mPress Enter to continue with DCI ref...\033[1;39m")

    dmm.switch_dci()
    dmm.inst.write("RANGE 100E-6")

    mfc.inst.write("R8\n") # set range 2mA
    mfc.dcv(0.100000)
    time.sleep(3)
    val = 0.0

    time.sleep(4)
    dmm.get_data()
    time.sleep(4)
    dmm.get_data()
    mfc_delta = (dmm_val * 1E3) - 0.100000
    print ("MFC error = %5.3f" % (((dmm_val / 100E-6) - 1) * 1E6) )
    b.write ("MFC error = %5.3f ppm \r\n" % (((dmm_val / 100E-6) - 1) * 1E6) )
    mfc.dcv(0.100000 - mfc_delta)
    time.sleep(1)

    for samples in range (0, 6):
        dmm.get_data()
        if (samples > 2):
            #print ("REF DCI %d : %.9E DCV" % (samples, dmm_val))
	    val += dmm_val
    refv = val / 3

    ref_dci_set[0] = (0.100000 - mfc_delta)
    ref_dci_meas[0] = refv
    print ("Measured REF DCI %d : %.9E ADC, stored DCI = %f" % (samples, ref_dci_meas[0], ref_dci_set[0]))
    
    ############### 1mA

    dmm.switch_dci()
    dmm.inst.write("RANGE 1E-3")

    mfc.inst.write("R8\n") # set range 2mA
    mfc.dcv(1.000000)
    time.sleep(3)
    val = 0.0

    time.sleep(4)
    dmm.get_data()
    time.sleep(4)
    dmm.get_data()
    mfc_delta = (dmm_val * 1E3) - 1.00000
    print ("MFC error = %5.3f" % (((dmm_val / 1E-3) - 1) * 1E6) )
    b.write ("MFC error = %5.3f ppm \r\n" % (((dmm_val / 1E-3) - 1) * 1E6) )
    mfc.dcv(1.000000 - mfc_delta)
    time.sleep(1)

    for samples in range (0, 6):
        dmm.get_data()
        if (samples > 2):
            #print ("REF DCI %d : %.9E DCV" % (samples, dmm_val))
	    val += dmm_val
    refv = val / 3

    ref_dci_set[1] = (1.00000 - mfc_delta)
    ref_dci_meas[1] = refv
    print ("Measured REF DCI %d : %.9E ADC, stored DCI = %f" % (samples, ref_dci_meas[1], ref_dci_set[1]))

    ############### 10mA

    dmm.switch_dci()
    dmm.inst.write("RANGE 1E-2")

    mfc.inst.write("R9\n") # set range 2mA
    mfc.dcv(10.000000)
    time.sleep(3)
    val = 0.0

    time.sleep(4)
    dmm.get_data()
    time.sleep(4)
    dmm.get_data()
    mfc_delta = (dmm_val * 1E3) - 10.00000
    print ("MFC error = %5.3f" % (((dmm_val / 10E-3) - 1) * 1E6) )
    b.write ("MFC error = %5.3f ppm \r\n" % (((dmm_val / 10E-3) - 1) * 1E6) )
    mfc.dcv(10.000000 - mfc_delta)
    time.sleep(1)

    for samples in range (0, 6):
        dmm.get_data()
        if (samples > 2):
            #print ("REF DCI %d : %.9E DCV" % (samples, dmm_val))
	    val += dmm_val
    refv = val / 3

    ref_dci_set[2] = (10.00000 - mfc_delta)
    ref_dci_meas[2] = refv
    print ("Measured REF DCI %d : %.9E ADC, stored DCI = %f" % (samples, ref_dci_meas[2], ref_dci_set[2]))

    ############### 100mA

    dmm.switch_dci()
    dmm.inst.write("RANGE 1E-1")

    mfc.inst.write("R10\n") # set range 2mA
    mfc.dcv(100.000000)
    time.sleep(3)
    val = 0.0

    time.sleep(4)
    dmm.get_data()
    time.sleep(4)
    dmm.get_data()
    mfc_delta = (dmm_val * 1E3) - 100.00000
    print ("MFC error = %5.3f" % (((dmm_val / 100E-3) - 1) * 1E6) )
    b.write ("MFC error = %5.3f ppm \r\n" % (((dmm_val / 100E-3) - 1) * 1E6) )
    mfc.dcv(100.000000 - mfc_delta)
    time.sleep(1)

    for samples in range (0, 6):
        dmm.get_data()
        if (samples > 2):
            #print ("REF DCI %d : %.9E DCV" % (samples, dmm_val))
	    val += dmm_val
    refv = val / 3

    ref_dci_set[3] = (100.00000 - mfc_delta)
    ref_dci_meas[3] = refv
    print ("Measured REF DCI %d : %.9E ADC, stored DCI = %f" % (samples, ref_dci_meas[3], ref_dci_set[3]))

    ############### 1000mA

    mfc.dcv(0.000000)

    dmm.switch_dci()
    dmm.inst.write("RANGE 1.0")

    mfc.inst.write("R11\n") # set range 2mA
    time.sleep(3)
    mfc.dcv(1.000000)
    val = 0.0

    time.sleep(4)
    dmm.get_data()
    time.sleep(4)
    dmm.get_data()
    mfc_delta = (dmm_val) - 1.00000
    print ("MFC error = %5.3f" % (((dmm_val / 1.0) - 1) * 1E6) )
    b.write ("MFC error = %5.3f ppm \r\n" % (((dmm_val / 1.0) - 1) * 1E6) )
    mfc.dcv(1.000000 - mfc_delta)
    time.sleep(1)

    for samples in range (0, 6):
        dmm.get_data()
        if (samples > 2):
            #print ("REF DCI %d : %.9E DCV" % (samples, dmm_val))
	    val += dmm_val
    refv = val / 3

    ref_dci_set[4] = (1.00000 - mfc_delta)
    ref_dci_meas[4] = refv
    print ("Measured REF DCI %d : %.9E ADC, stored DCI = %f" % (samples, ref_dci_meas[4], ref_dci_set[4]))
    #samples = 0
    #dval = 0

    #for samples in range (0, 6):
#	kei2000.write(":READ?")
#    	val = float(kei2000.read())
#        if (samples > 2):
#            #print ("Test 2000 DCV %d : %.9E DCV" % (samples, val))
#	    dval += val
#    dval = dval / 3

    #print ("Measured 2000 DCV %d : %.9E VDC" % (samples, dval))
    #print ("Delta REF/DUT = %.9E , %5.3f ppm" % ( (refv - dval), (((refv / dval) - 1) * 1E6) ) )
    #b.write ("HP 3458A DCI %.9E ADC, K2000 DUT %.9E ADC, Delta %5.3f ppm\r\n" % (refv, dval, (((refv / dval) - 1) * 1E6) ) )

    mfc.dcv(0.000000)

    quit()

